/*
 * Copyright (c) 2005 Ariff Abdullah 
 *        <skywizard@MyBSD.org.my> All rights reserved.
 *
 * Derived from vchan.c
 *
 * Copyright (c) 2001 Cameron Grant
 *        <cg@freebsd.org> All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                                         *
 * This new implementation is fully dedicated in memory of Cameron Grant,  *
 * the creator of magnificent, highly addictive feeder infrastructure.     *
 *                                                                         *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 */

#include <dev/sound/pcm/sound.h>
#include <dev/sound/pcm/vchan.h>
#include "feeder_if.h"

SND_DECLARE_FILE("$FreeBSD$");

struct slaveinfo {
	uint32_t spd, fmt, blksz, bps, run;
	struct pcm_channel *channel, *parent;
};

static int
feed_slave(struct pcm_feeder *f, struct pcm_channel *c, uint8_t *b,
		uint32_t count, void *source)
{
	struct pcmchan_children *cce;
	struct pcm_channel *ch;
	int cnt = 0;

	cce = SLIST_FIRST(&c->children);
	if (!cce || !cce->channel)
		return 0;
	ch = cce->channel;
	CHN_LOCK(ch);
	if (ch->flags & CHN_F_TRIGGERED) {
		if (ch->flags & CHN_F_MAPPED)
			sndbuf_acquire(ch->bufsoft, NULL, sndbuf_getfree(ch->bufsoft));
		cnt = FEEDER_FEED(ch->feeder, ch, b, count, ch->bufsoft);
	}
	CHN_UNLOCK(ch);
	return cnt;
}

static struct pcm_feederdesc feeder_slave_desc[] = {
	{FEEDER_SLAVE, AFMT_U8, AFMT_U8, 0},
	{FEEDER_SLAVE, AFMT_U8|AFMT_STEREO, AFMT_U8|AFMT_STEREO, 0},
	{FEEDER_SLAVE, AFMT_S8, AFMT_S8, 0},
	{FEEDER_SLAVE, AFMT_S8|AFMT_STEREO, AFMT_S8|AFMT_STEREO, 0},
	{FEEDER_SLAVE, AFMT_A_LAW, AFMT_A_LAW, 0},
	{FEEDER_SLAVE, AFMT_A_LAW|AFMT_STEREO, AFMT_A_LAW|AFMT_STEREO, 0},
	{FEEDER_SLAVE, AFMT_MU_LAW, AFMT_MU_LAW, 0},
	{FEEDER_SLAVE, AFMT_MU_LAW|AFMT_STEREO, AFMT_MU_LAW|AFMT_STEREO, 0},
	{FEEDER_SLAVE, AFMT_U16_LE, AFMT_U16_LE, 0},
	{FEEDER_SLAVE, AFMT_U16_LE|AFMT_STEREO, AFMT_U16_LE|AFMT_STEREO, 0},
	{FEEDER_SLAVE, AFMT_S16_LE, AFMT_S16_LE, 0},
	{FEEDER_SLAVE, AFMT_S16_LE|AFMT_STEREO, AFMT_S16_LE|AFMT_STEREO, 0},
	{FEEDER_SLAVE, AFMT_U16_BE, AFMT_U16_BE, 0},
	{FEEDER_SLAVE, AFMT_U16_BE|AFMT_STEREO, AFMT_U16_BE|AFMT_STEREO, 0},
	{FEEDER_SLAVE, AFMT_S16_BE, AFMT_S16_BE, 0},
	{FEEDER_SLAVE, AFMT_S16_BE|AFMT_STEREO, AFMT_S16_BE|AFMT_STEREO, 0},
	{FEEDER_SLAVE, AFMT_U24_LE, AFMT_U24_LE, 0},
	{FEEDER_SLAVE, AFMT_U24_LE|AFMT_STEREO, AFMT_U24_LE|AFMT_STEREO, 0},
	{FEEDER_SLAVE, AFMT_S24_LE, AFMT_S24_LE, 0},
	{FEEDER_SLAVE, AFMT_S24_LE|AFMT_STEREO, AFMT_S24_LE|AFMT_STEREO, 0},
	{FEEDER_SLAVE, AFMT_U24_BE, AFMT_U24_BE, 0},
	{FEEDER_SLAVE, AFMT_U24_BE|AFMT_STEREO, AFMT_U24_BE|AFMT_STEREO, 0},
	{FEEDER_SLAVE, AFMT_S24_BE, AFMT_S24_BE, 0},
	{FEEDER_SLAVE, AFMT_S24_BE|AFMT_STEREO, AFMT_S24_BE|AFMT_STEREO, 0},
	{FEEDER_SLAVE, AFMT_U32_LE, AFMT_U32_LE, 0},
	{FEEDER_SLAVE, AFMT_U32_LE|AFMT_STEREO, AFMT_U32_LE|AFMT_STEREO, 0},
	{FEEDER_SLAVE, AFMT_S32_LE, AFMT_S32_LE, 0},
	{FEEDER_SLAVE, AFMT_S32_LE|AFMT_STEREO, AFMT_S32_LE|AFMT_STEREO, 0},
	{FEEDER_SLAVE, AFMT_U32_BE, AFMT_U32_BE, 0},
	{FEEDER_SLAVE, AFMT_U32_BE|AFMT_STEREO, AFMT_U32_BE|AFMT_STEREO, 0},
	{FEEDER_SLAVE, AFMT_S32_BE, AFMT_S32_BE, 0},
	{FEEDER_SLAVE, AFMT_S32_BE|AFMT_STEREO, AFMT_S32_BE|AFMT_STEREO, 0},
	{0, 0, 0, 0},
};
static kobj_method_t feeder_slave_methods[] = {
    	KOBJMETHOD(feeder_feed, feed_slave),
	{0, 0}
};
FEEDER_DECLARE(feeder_slave, 2, NULL);

static void *
slave_init(kobj_t obj, void *devinfo, struct snd_dbuf *b,
			struct pcm_channel *c, int dir)
{
	struct slaveinfo *ch;
	struct pcm_channel *parent = devinfo;

	KASSERT(dir == PCMDIR_PLAY ||
			dir == PCMDIR_REC, ("slave_init: bad direction"));
	ch = malloc(sizeof(*ch), M_DEVBUF, M_WAITOK|M_ZERO);
	if (!ch)
		return NULL;
	ch->parent = parent;
	ch->channel = c;
	ch->fmt = AFMT_U8;
	ch->spd = DSP_DEFAULT_SPEED;
	ch->blksz = 2048;
	c->flags |= CHN_F_SLAVE;
	return ch;
}

static int
slave_free(kobj_t obj, void *data)
{
	return 0;
}

static int
slave_setformat(kobj_t obj, void *data, uint32_t format)
{
	struct slaveinfo *ch = data;
	struct pcm_channel *parent = ch->parent;
	struct pcm_channel *channel = ch->channel;
	struct pcm_feeder *f;
	int r;

	CHN_UNLOCK(channel);
	CHN_LOCK(parent);
	r = CHANNEL_SETFORMAT(parent->methods, parent->devinfo, format);
	if (r == 0) {
		parent->format = format;
		sndbuf_setfmt(parent->bufhard, format);
		sndbuf_setfmt(parent->bufsoft, format);
		f = parent->feeder;
		if (f && f->desc && f->desc->type == FEEDER_SLAVE) {
			f->desc->in = format;
			f->desc->out = format;
		}
	}
	CHN_UNLOCK(parent);
	CHN_LOCK(channel);
	if (r == 0) {
		ch->fmt = format;
		ch->bps = 1;
		if (ch->fmt & AFMT_STEREO)
			ch->bps <<= 1;
		if (ch->fmt & AFMT_16BIT)
			ch->bps <<= 1;
		else if (ch->fmt & AFMT_24BIT)
			ch->bps *= 3;
		else if (ch->fmt & AFMT_32BIT)
			ch->bps <<= 2;
		sndbuf_setfmt(channel->bufsoft, format);
		sndbuf_setfmt(channel->bufhard, format);
	}
	return r;
}

static int
slave_setspeed(kobj_t obj, void *data, uint32_t speed)
{
	struct slaveinfo *ch = data;
	struct pcm_channel *parent = ch->parent;
	struct pcm_channel *channel = ch->channel;
	int r, hwspd = speed;

   	CHN_UNLOCK(channel);
	CHN_LOCK(parent);
	RANGE(hwspd, chn_getcaps(parent)->minspeed, chn_getcaps(parent)->maxspeed);
	r = CHANNEL_SETSPEED(parent->methods, parent->devinfo, hwspd);
	parent->speed = r;
	sndbuf_setspd(parent->bufhard, r);
	sndbuf_setspd(parent->bufsoft, r);
	CHN_UNLOCK(parent);
   	CHN_LOCK(channel);
	ch->spd = r;
	return ch->spd;
}

static int
slave_setblocksize(kobj_t obj, void *data, uint32_t blocksize)
{
	struct slaveinfo *ch = data;
	struct pcm_channel *channel = ch->channel;
	struct pcm_channel *parent = ch->parent;
	int prate, crate;

	ch->blksz = blocksize;
	sndbuf_setblksz(channel->bufhard, blocksize);
	chn_notify(parent, CHN_N_BLOCKSIZE);
   	CHN_LOCK(parent);
	crate = ch->spd * ch->bps;
	prate = sndbuf_getspd(parent->bufhard) * sndbuf_getbps(parent->bufhard);
	blocksize = sndbuf_getblksz(parent->bufhard);
   	CHN_UNLOCK(parent);
	blocksize *= prate;
	blocksize /= crate;
	return blocksize;
}

static int
slave_trigger(kobj_t obj, void *data, int go)
{
	struct slaveinfo *ch = data;
	struct pcm_channel *parent = ch->parent;
	struct pcm_channel *channel = ch->channel;

	if (go == PCMTRIG_EMLDMAWR || go == PCMTRIG_EMLDMARD)
		return 0;
	ch->run = (go == PCMTRIG_START) ? 1 : 0;
   	CHN_UNLOCK(channel);
	chn_notify(parent, CHN_N_TRIGGER);
   	CHN_LOCK(channel);
	return 0;
}

static struct pcmchan_caps *
slave_getcaps(kobj_t obj, void *data)
{
	struct slaveinfo *ch = data;
	struct pcm_channel *channel = ch->channel;
	struct pcm_channel *parent = ch->parent;
	struct pcmchan_caps *caps;

	CHN_UNLOCK(channel);
	CHN_LOCK(parent);
	caps = chn_getcaps(parent);
	CHN_UNLOCK(parent);
	CHN_LOCK(channel);
	return caps;
}

static kobj_method_t slave_methods[] = {
    	KOBJMETHOD(channel_init,		slave_init),
    	KOBJMETHOD(channel_free,		slave_free),
    	KOBJMETHOD(channel_setformat,		slave_setformat),
    	KOBJMETHOD(channel_setspeed,		slave_setspeed),
    	KOBJMETHOD(channel_setblocksize,	slave_setblocksize),
    	KOBJMETHOD(channel_trigger,		slave_trigger),
    	KOBJMETHOD(channel_getcaps,		slave_getcaps),
	{0, 0}
};
CHANNEL_DECLARE(slave);

int
slave_create(struct pcm_channel *parent)
{
    	struct snddev_info *d = parent->parentsnddev;
	struct pcmchan_children *pce;
	struct pcm_channel *child;
	struct pcmchan_caps *parent_caps;
	int err;

	if (!(parent->flags & CHN_F_BUSY) || !SLIST_EMPTY(&parent->children))
		return EBUSY;

	parent_caps = chn_getcaps(parent);
	if (parent_caps == NULL || parent_caps->fmtlist == NULL)
		return EINVAL;

   	CHN_UNLOCK(parent);

	pce = malloc(sizeof(*pce), M_DEVBUF, M_WAITOK | M_ZERO);
	if (!pce) {
   		CHN_LOCK(parent);
		return ENOMEM;
	}

	child = pcm_chn_create(d, parent, &slave_class, PCMDIR_SLAVE, parent);
	if (!child) {
		free(pce, M_DEVBUF);
   		CHN_LOCK(parent);
		return ENODEV;
	}
	pce->channel = child;

	err = pcm_chn_add(d, child);
	if (err) {
		pcm_chn_destroy(child);
		free(pce, M_DEVBUF);
		CHN_LOCK(parent);
		return err;
	}

	CHN_LOCK(parent);
	SLIST_INSERT_HEAD(&parent->children, pce, link);
	parent->flags |= CHN_F_HAS_SLAVE;

	err = chn_reset(parent, parent_caps->fmtlist[0]);
	if (!err)
		err = chn_setspeed(parent, parent_caps->minspeed);
	if (err) {
		SLIST_REMOVE(&parent->children, pce, pcmchan_children, link);
		parent->flags &= ~CHN_F_HAS_SLAVE;
		CHN_UNLOCK(parent);
		free(pce, M_DEVBUF);
		pcm_chn_remove(d, child);
		pcm_chn_destroy(child);
		CHN_LOCK(parent);
		return err;
	}

	return 0;
}

int
slave_destroy(struct pcm_channel *c)
{
	struct pcm_channel *parent = c->parentchannel;
    	struct snddev_info *d = parent->parentsnddev;
	struct pcmchan_children *pce;
	struct snddev_channel *sce;
	int err, last;

	CHN_LOCK(parent);
	if (!(parent->flags & CHN_F_BUSY)) {
		CHN_UNLOCK(parent);
		return EBUSY;
	}
	if (SLIST_EMPTY(&parent->children)) {
		CHN_UNLOCK(parent);
		return EINVAL;
	}
	SLIST_FOREACH(pce, &parent->children, link) {
		if (pce->channel == c)
			goto gotchslave;
	}
	CHN_UNLOCK(parent);
	return EINVAL;
gotchslave:
	SLIST_FOREACH(sce, &d->channels, link) {
		if (sce->channel == c) {
			if (sce->dsp_devt)
				destroy_dev(sce->dsp_devt);
			if (sce->dspW_devt)
				destroy_dev(sce->dspW_devt);
			if (sce->audio_devt)
				destroy_dev(sce->audio_devt);
			if (sce->dspr_devt)
				destroy_dev(sce->dspr_devt);
			break;
		}
	}
	SLIST_REMOVE(&parent->children, pce, pcmchan_children, link);
	free(pce, M_DEVBUF);

	last = SLIST_EMPTY(&parent->children);
	if (last) {
		parent->flags &= ~CHN_F_BUSY;
		parent->flags &= ~CHN_F_HAS_SLAVE;
	}
	err = pcm_chn_remove(d, c);
	CHN_UNLOCK(parent);
	if (!err)
		err = pcm_chn_destroy(c);
#if 0
	if (!err && last) {
		CHN_LOCK(parent);
		chn_reset(parent, chn_getcaps(parent)->fmtlist[0]);
		chn_setspeed(parent, chn_getcaps(parent)->minspeed);
		CHN_UNLOCK(parent);
	}
#endif
	return err;
}
